package com.deltadna.android.sdk;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;
import java.util.UUID;

import org.json.JSONException;
import org.json.JSONObject;

import android.app.Application;
import android.os.Bundle;
import android.util.Log;

import com.deltadna.android.sdk.helpers.ClientInfo;
import com.deltadna.android.sdk.helpers.EngageArchive;
import com.deltadna.android.sdk.helpers.EventStore;
import com.deltadna.android.sdk.helpers.NotStartedException;
import com.deltadna.android.sdk.helpers.PlayerPrefs;
import com.deltadna.android.sdk.helpers.Settings;
import com.deltadna.android.sdk.helpers.Utils;
import com.deltadna.android.sdk.net.DDNANetManager;
import com.deltadna.android.sdk.net.DDNANetReq;
import com.deltadna.android.sdk.net.DDNANetReqCb;
/**
 * Singleton SDK class, use DDNA.inst() to obtain the singleton.
 * 
 * Must be initialised once before use using DDNA.inst().init(android.app.Application).
 * 
 * DDNA.inst().startSDK(String envKey, String collectURL, String engageURL, String userID) must be called to
 * begin collection and DDNA.inst().stopSDK() called to stop collection (while application is paused).
 * 
 * com.deltadna.android.sdk.helpers.Settings and com.deltadna.android.sdk.helpers.PlayerPrefs can be used to
 * modify default behaviour.
 */
public class DDNA{
	static private DDNA smInst = null;

	static public final String LOGTAG = "DeltaDNA";

	static public final String PF_KEY_USER_ID = "DDSDK_USER_ID";
	static public final String PF_KEY_FIRST_RUN = "DDSDK_FIRST_RUN";
	static public final String PF_KEY_HASH_SECRET = "DDSDK_HASH_SECRET";
	static public final String PF_KEY_CLIENT_VERSION = "DDSDK_CLIENT_VERSION";
	static public final String PF_KEY_PUSH_NOTIFICATION_TOKEN = "DDSDK_PUSH_NOTIFICATION_TOKEN";
	static public final String PF_KEY_ANDROID_REGISTRATION_ID = "DDSDK_ANDROID_REGISTRATION_ID";
	static private final String PF_KEY_STORAGE_PATH = "DDSDK_STORAGE_PATH";

	static public final String EV_KEY_NAME = "eventName";
	static public final String EV_KEY_USER_ID = "userID";
	static public final String EV_KEY_SESSION_ID = "sessionID";
	static public final String EV_KEY_TIMESTAMP = "eventTimestamp";
	static public final String EV_KEY_PARAMS = "eventParams";

	static public final String EP_KEY_PLATFORM = "platform";
	static public final String EP_KEY_SDK_VERSION = "sdkVersion";

	static public final String EXTRA_DDNAVALUE = "ddnaValue";
	static public final String EXTRA_DDNAPARAMS = "ddnaParams";
	static public final String EXTRA_IMGMESSAGE = "imgmessage";
    static public final String EXTRA_PUSH_NOTIFICATION_ID = "notificationId";
    static public final String EXTRA_PUSH_NOTIFICATION_NAME = "notificationName";
    static public final String EXTRA_PUSH_NOTIFICATION_LAUNCH = "notificationLaunch";
	static public final int RESULTCODE_IMGMESSAGE = 0x00001;

	static public final String AUTO_GENERATED_USER_ID = null;

	private Application mApp = null;

	private EventStore mEventStore = null;
	private EngageArchive mEngageArchive = null;

    private Settings mSettings = null;
	private TransactionBuilder mTransaction = null;

	private String mEnvironmentKey = null;
	private String mCollectURL = null;
	private String mEngageURL = null;
	private String mSessionID = null;
	private String mPlatform = null;
	private boolean mIsStarted = false;

	/**
	 * Private constructor to force a singleton pattern. Use DDNA.inst() to obtain
	 * the singleton.
	 */
	private DDNA(){
        mSettings = new Settings();
    }
	/**
	 * Convenience method to return the current Application object, needs to be set by
	 * the user using init().
	 * 
	 * @return The current Application object, null of not set.
	 */
	public Application app(){
		return mApp;
	}
	/**
	 * Gets the current EventStore object.
	 * 
	 * @return The EventStore object.
	 */
	public EventStore eventStore(){
		return mEventStore;
	}
	/**
	 * Gets the Engage archive of previous Engage call responses.
	 * 
	 * @return The current EngageArchive.
	 */
	public EngageArchive engageArchive(){
		return mEngageArchive;
	}
	/**
	 * Changes the session ID for the current User.
	 */
	public void newSession(){
		mSessionID = UUID.randomUUID().toString();
	}
	/**
	 * Helper for building common transaction type events.
	 * 
	 * @return The transaction builder.
	 */
    public Settings settings() {
        return mSettings;
    }

	public TransactionBuilder transaction(){
		return mTransaction;
	}
	/**
	 * Gets the unique environment key for this game environment.
	 * 
	 * @return The environment key if set, null on error.
	 */
	public String environmentKey(){
		return mEnvironmentKey;
	}
	/**
	 * Gets the Collect URL for this game.
	 * 
	 * @return The Collect URL if set, null on error.
	 */
	public String collectURL(){
		return mCollectURL;
	}
	/**
	 * Gets the Engage URL for this game.
	 * 
	 * @return The Engage URL if set, null on error.
	 */
	public String engageURL(){
		return mEngageURL;
	}
	/**
	 * Gets the current session ID.
	 * 
	 * @return The current session id, null if none exists yet.
	 */
	public String sessionID(){
		return mSessionID;
	}
	/**
	 * Gets the platform.
	 * 
	 * @return
	 */
	public String platform(){
		return mPlatform;
	}
	/**
	 * Gets the user ID.
	 * 
	 * @return
	 */
	public String userID(){
		String v = PlayerPrefs.getString(PF_KEY_USER_ID, null);
		if (Utils.isNullOrEmpty(v)){
			Log.d(DDNA.LOGTAG, "No existing User ID found.");
			return null;
		} 
		return v;
	}
	/**
	 * 
	 * @param userID
	 */
	private void setUserID(String userID){
		String currentUserId = PlayerPrefs.getString(PF_KEY_USER_ID, null);
		if(Utils.isNullOrEmpty(userID)){
			if(Utils.isNullOrEmpty(currentUserId)){
				Log.d(DDNA.LOGTAG, "Creating a new user id for player");
				PlayerPrefs.putString(PF_KEY_USER_ID, UUID.randomUUID().toString());
				PlayerPrefs.save();
			}

		}else{
			if(!Utils.isNullOrEmpty(currentUserId)){
				if(!currentUserId.equals(userID)){
					PlayerPrefs.remove(PF_KEY_FIRST_RUN);
					PlayerPrefs.putString(PF_KEY_USER_ID, userID);
					PlayerPrefs.save();
				}
			}else{
				PlayerPrefs.remove(PF_KEY_FIRST_RUN);
				PlayerPrefs.putString(PF_KEY_USER_ID, userID);
				PlayerPrefs.save();
			}
		}
	}
	/**
	 * Gets a value indicating whether startSDK() has been called.
	 * 
	 * @return TRUE if the SDK is running, FALSE otherwise.
	 */
	public boolean started(){
		return mIsStarted;
	}
	/**
	 * Gets the hash secret key.
	 * 
	 * @return The hash secret string, null if not enabled or on error.
	 */
	public String hashSecret(){
		String v = PlayerPrefs.getString(PF_KEY_HASH_SECRET, null);
		if (Utils.isNullOrEmpty(v)){
			return null;
		}
		return v;
	}
	/**
	 * To enable hashing of your event and engage data, set this value to your
	 * unique hash secret.  You must also enable hashing for the environment.
	 * To disable hashing set it to null, which is the default.  This must be 
	 * set before calling startSDK().
	 * 
	 * @param secret The new secret value.
	 */
	public void setHashSecret(String secret){
		PlayerPrefs.putString(DDNA.PF_KEY_HASH_SECRET, secret);
		PlayerPrefs.save();
	}
	/**
	 * Gets the client version string.
	 * 
	 * @return The client version string, null if not set.
	 */
	public String clientVersion(){
		String v = PlayerPrefs.getString(PF_KEY_CLIENT_VERSION, null);
		if (Utils.isNullOrEmpty(v)){
			Log.w(DDNA.LOGTAG, "No client game version set.");
			return null;
		}
		return v;
	}
	/**
	 * A version String for your game that will be reported to us.  This must
	 * be set before calling startSDK().
	 * 
	 * @param version The version string.
	 */
	public void setClientVersion(String version){
		PlayerPrefs.putString(DDNA.PF_KEY_CLIENT_VERSION, version);
		PlayerPrefs.save();
	}
	/**
	 * Gets the Android registration ID.
	 * 
	 * @return The android registration id, null if not set.
	 */
	public String androidRegistrationID(){
		String v = PlayerPrefs.getString(PF_KEY_ANDROID_REGISTRATION_ID, null);
		if (Utils.isNullOrEmpty(v)){
			if (ClientInfo.platform().contains("ANDROID")){
				Log.w(DDNA.LOGTAG, "No Android registration id set, sending push notifications to Android devices will be unavailable.");					
			}
			return null;
		}	
		return v;
	}
	/**
	 * The Android registration ID that is associated with this device if it's running
	 * on the Android platform. This must be set before calling startSDK().
	 * 
	 * @param id The Android registration id.
	 */
	public void setAndroidRegistrationID(String id){
		PlayerPrefs.putString(DDNA.PF_KEY_ANDROID_REGISTRATION_ID, id);
		PlayerPrefs.save();
	}
	/**
	 * Gets a formatted timestamp string.
	 * 
	 * @return The timestamp string.
	 */
	private String getCurrentTimestamp(){
		String result = null;

		final SimpleDateFormat sdf = new SimpleDateFormat(Settings.EVENT_TIMESTAMP_FORMAT, Locale.US);
		sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
		result = sdf.format(new Date());

		return result;
	}
	/**
	 * The storage directory the DeltaDNA SDK will use for storing temp data.
	 * 
	 * @return The directory path.
	 */
	public String storageDir(){
		String result = PlayerPrefs.getString(PF_KEY_STORAGE_PATH, null);

		if(result == null){
			File dir = mApp.getApplicationContext().getExternalFilesDir(null);
			if(dir != null){
				result = dir.getAbsolutePath();
			}else{
				result = "/";
			}
		}

		return result;
	}
	/**
	 * Required to tell the SDK what Application it belongs to.
	 * 
	 * @param app The current application object.
	 */
	public void init(Application app){
		mApp = app;

		PlayerPrefs.init(app);
	}
	/**
	 * Starts the DDNA.  Call before sending events or making engagements.
	 * 
	 * @param envKey The unique environment key for this game environment.
	 * @param collectURL The Collect URL for this game.
	 * @param engageURL The Engage URL for this game.
	 * @param userID The user id for the player, if set to AUTO_GENERATED_USER_ID we create one for you.
	 */
	public void startSDK(String envKey, String collectURL, String engageURL, String userID){
		if(!mIsStarted && (mApp != null)){
			if(mTransaction == null){
				mTransaction = new TransactionBuilder();

				mEventStore = new EventStore(Settings.EVENT_STORAGE_PATH.replace("{persistent_path}", storageDir()), 
						mSettings.debugMode());


				mEngageArchive = new EngageArchive(Settings.ENGAGE_STORAGE_PATH.replace("{persistent_path}", storageDir()));

				setUserID(userID);

				mEnvironmentKey = envKey;
				mCollectURL = validateURL(collectURL);
				mEngageURL = validateURL(engageURL);
				mPlatform = ClientInfo.platform();

				newSession();
			}

			mIsStarted = true;

			// must do this once we're mInitialised
			triggerDefaultEvents();

			// Setup automated event uploads
			if (mSettings.backgroundEventUpload()){
				DDNANetManager.inst().start(mSettings.backgroundEventUploadStartDelaySeconds(), mSettings.backgroundEventUploadRepeatRateSeconds());
			}
			Log.d(DDNA.LOGTAG, "Started SDK");
		}
	}
	/**
	 * Sends a 'gameEnded' event to Collect, disables background uploads.
	 * 
	 * @throws NotStartedException
	 */
	public void stopSDK() throws NotStartedException{
		if (mIsStarted) {
			Log.d(DDNA.LOGTAG, "Stopping SDK");
			DDNANetManager.inst().stop();

			recordEvent("gameEnded");

			DDNANetManager.inst().uploadEvents(null);

			if (mEngageArchive != null) {
				mEngageArchive.save();
			}
			mIsStarted = false;
			Log.d(DDNA.LOGTAG, "Stopped SDK");
		} else {
			Log.d(DDNA.LOGTAG, "SDK not running");
		}
	}
	/**
	 * Clears the persistent data such as user id.  Useful for testing purposes.
	 */
	public void clearPersistentData(){
		// PlayerPrefs
		PlayerPrefs.remove(PF_KEY_USER_ID);
		PlayerPrefs.remove(PF_KEY_FIRST_RUN);
		PlayerPrefs.remove(PF_KEY_HASH_SECRET);
		PlayerPrefs.remove(PF_KEY_CLIENT_VERSION);
		PlayerPrefs.remove(PF_KEY_PUSH_NOTIFICATION_TOKEN);
		PlayerPrefs.remove(PF_KEY_ANDROID_REGISTRATION_ID);	
		PlayerPrefs.save();

		if(mEventStore != null){
			mEventStore.clear();
		}
		if(mEngageArchive != null){
			mEngageArchive.clear();
		}
	}
	/**
	 * Sends an event to Collect, with no additional event parameters.
	 * 
	 * @param eventName Name of the event.
	 * @throws NotStartedException
	 */
	public void recordEvent(String eventName) throws NotStartedException{
		recordEvent(eventName, new JSONObject());
	}
	/**
	 * Sends an event to Collect.
	 * 
	 * @param eventName Name of the event schema.
	 * @param eventParams An EventBuilder that describes the event params for the event.
	 * @throws NotStartedException
	 */
	public void recordEvent(String eventName, EventBuilder eventParams) throws NotStartedException{
		recordEvent(eventName, eventParams == null ? new JSONObject() : eventParams.toJSON());
	}
	/**
	 * Sends an event to Collect.
	 * 
	 * @param eventName Name of the event schema.
	 * @param eventParams Event parameters for the event.
	 * @throws NotStartedException
	 */
	public void recordEvent(String eventName, JSONObject eventParams) throws NotStartedException{
		if (!mIsStarted) {
			throw new NotStartedException("You must first start the SDK via the startSDK method");
		}

		// the header for every event is eventName, userID, sessionID and timestamp
		JSONObject eventRecord = new JSONObject();
		try{
			eventRecord.put(EV_KEY_NAME, eventName);
			eventRecord.put(EV_KEY_USER_ID, userID());
			eventRecord.put(EV_KEY_SESSION_ID, mSessionID);
			eventRecord.put(EV_KEY_TIMESTAMP, getCurrentTimestamp());

			// every template should support sdkVersion and platform in it's event params
			if (!eventParams.has(EP_KEY_PLATFORM)){
				eventParams.put(EP_KEY_PLATFORM, mPlatform);
			}

			if (!eventParams.has(EP_KEY_SDK_VERSION)){
				eventParams.put(EP_KEY_SDK_VERSION, Settings.SDK_VERSION);
			}

			eventRecord.put(EV_KEY_PARAMS, eventParams);

		}catch(JSONException e){}

		if (Utils.isNullOrEmpty(userID())){

		}else if (!mEventStore.push(eventRecord.toString())){
			Log.w(DDNA.LOGTAG, "Event Store full, unable to handle event");
		}
	}

    /**
     * Record when a push notification is received.
     * @param extras A bundle containing the notification id, name and flag to indicate if the app
     *               was launched by the notification.
     */
    public void recordPushNotification(Bundle extras) throws NotStartedException {
        if (!extras.isEmpty()) {
            int id = extras.getInt(EXTRA_PUSH_NOTIFICATION_ID, 0);
            String name = extras.getString(EXTRA_PUSH_NOTIFICATION_NAME) != null ? extras.getString(EXTRA_PUSH_NOTIFICATION_NAME) : "UNKNOWN"; // Support API 9
            boolean launch = extras.getBoolean(EXTRA_PUSH_NOTIFICATION_LAUNCH, true);

            EventBuilder notificationParams = new EventBuilder();
            notificationParams.addParam(EXTRA_PUSH_NOTIFICATION_ID, id);
            notificationParams.addParam(EXTRA_PUSH_NOTIFICATION_NAME, name);
            notificationParams.addParam(EXTRA_PUSH_NOTIFICATION_LAUNCH, launch);

            this.recordEvent("notificationOpened", notificationParams);
        }
    }

    /**
     * Sends waiting events to our platform.  This is usually called automatically, but automation
     * can be disabled, in which case you are responsible for calling upload at suitable points in
     * the game flow.
     * @throws NotStartedException
     */
    public void upload() throws NotStartedException {
        DDNANetManager.inst().uploadEvents(null);
    }

	/**
	 * Fires the default events, will only be called from startSDK().
	 */
	private void triggerDefaultEvents(){
		try{
			if(mSettings.onFirstRunSendNewPlayerEvent() && PlayerPrefs.getInt(PF_KEY_FIRST_RUN, 1) > 0){
				Log.d(DDNA.LOGTAG, "Sending 'newPlayer' event");

				EventBuilder newPlayerParams = new EventBuilder()
				.addParam("userCountry", ClientInfo.countryCode());

				this.recordEvent("newPlayer", newPlayerParams);

				PlayerPrefs.putInt(PF_KEY_FIRST_RUN, 0);
				PlayerPrefs.save();
			}

			if(mSettings.onInitSendGameStartedEvent()){
				Log.d(DDNA.LOGTAG, "Sending 'gameStarted' event");

				EventBuilder gameStartedParams = new EventBuilder()
				.addParam("clientVersion", clientVersion())
				.addParam("androidRegistrationID", androidRegistrationID())
				.addParam("userLocale", ClientInfo.locale());

				this.recordEvent("gameStarted", gameStartedParams);
			}

			if(mSettings.onInitSendClientDeviceEvent()){
				Log.d(DDNA.LOGTAG, "Sending 'clientDevice' event");

				EventBuilder clientDeviceParams = new EventBuilder()
				.addParam("deviceName", ClientInfo.deviceName())
				.addParam("deviceType", ClientInfo.deviceType())
				.addParam("hardwareVersion", ClientInfo.deviceModel())
				.addParam("operatingSystem", ClientInfo.operatingSystem())
				.addParam("operatingSystemVersion", ClientInfo.operatingSystemVersion())
				.addParam("manufacturer", ClientInfo.manufacturer())
				.addParam("timezoneOffset", ClientInfo.timezoneOffset())
				.addParam("userLanguage", ClientInfo.languageCode());

				this.recordEvent("clientDevice", clientDeviceParams);
			}
			
		}catch(NotStartedException e){
			e.printStackTrace();
		}
	}
	/**
	 * Makes an Engage request.  The result of the engagement will be passed as a dictionary object to your callback method.
	 * 
	 * @param decisionPoint The decision point the request is for, must match the String in Portal.
	 * @param engageParams Additional parameters for the engagement.
	 * @param cb Optional user callback to notify on network events.
	 * @throws NotStartedException
	 */
	public DDNANetReq requestEngagement(String decisionPoint, JSONObject engageParams, DDNANetReqCb cb) throws NotStartedException{
		return DDNANetManager.inst().engageRequest(decisionPoint, engageParams, cb);
	}
	/**
	 * Requests an image message engagement.
	 * 
	 * @param decisionPoint The decision point the request is for, must match the String in Portal.
	 * @param engageParams Additional parameters for the engagement.
	 * @param cb Optional user callback to notify on network events.
	 * @throws NotStartedException
	 */
	public DDNANetReq requestImageMessage(String decisionPoint, JSONObject engageParams, DDNANetReqCb cb) throws NotStartedException{
		return DDNANetManager.inst().requestImageMessage(decisionPoint, engageParams, cb);
	}
	/**
	 * Gets the singleton instance of the SDK.
	 * 
	 * @return The singleton DDNA SDK object.
	 */
	static public DDNA inst(){
		if(smInst == null){
			smInst = new DDNA();
		}
		return smInst;
	}

    private String validateURL(String url) {
        if (!url.toLowerCase(Locale.getDefault()).startsWith("http://") && !url.toLowerCase(Locale.getDefault()).startsWith("https://")) {
            url = "http://" + url;
        }
        return url;
    }
}


