package com.deltadna.android.sdk;

import org.json.JSONException;

import com.deltadna.android.sdk.ImgMessage;

import android.annotation.SuppressLint;
import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.os.Bundle;
import android.util.AttributeSet;
import android.util.Log;
import android.view.MotionEvent;
import android.view.View;
import android.widget.FrameLayout;
import android.widget.FrameLayout.LayoutParams;
import android.widget.ProgressBar;
import android.widget.RelativeLayout;

/**
 * The popup Activity is intended as an default method to display the ImgMessage provided
 * by an Engage request.
 * 
 * To enable the use of this Activity it must be added to the AndroidManifest.xml :
 * 
 * <activity
 * 		android:name="com.deltadna.android.sdk.DDNAPopupActivity"
 * 		android:label="@string/app_name"
 * 		android:theme="@android:style/Theme.Light.NoTitleBar"
 * 		android:windowIsTranslucent="true"
 * 		android:windowBackground="@android:color/transparent">
 * 		<intent-filter>
 * 			<action android:name="android.intent.action.ACTION_DEFAULT" />
 * 
 * 			<category android:name="android.intent.category.SAMPLE_CODE" />
 * 		</intent-filter>
 * </activity>
 * 
 * May be invoked using:
 * 
 * Intent intent = new Intent(Activity.this, DDNAPopupActivity.class);
 * 
 * intent.putExtra(DDNA.EXTRA_IMGMESSAGE, req.responseData().toString());
 * 
 * Activity.this.startActivityForResult(intent, DDNA.RESULTCODE_IMGMESSAGE);
 * 
 * Will return a result to onActivityResult() :
 * 
 * 		if(requestCode == DDNA.RESULTCODE_IMGMESSAGE){
 *			if(resultCode == Activity.RESULT_OK){
 *				Bundle extras = data.getExtras();
 *
 *				String value = extras.getString(DDNA.EXTRA_DDNAVALUE);
 *				String params = extras.getString(DDNA.EXTRA_DDNAPARAMS);
 *
 *				// do something with the result
 *				showDialog("Image Message", "Now we are back in the game with the message values: " + value + " and params: " + params);
 *			}else if(resultCode == Activity.RESULT_CANCELED){
 *				// dont do something with the result
 *				showDialog("Image Message", "We have dismissed the message with no action.");
 *			}
 *		}
 */
public class DDNAPopupActivity extends Activity implements ImgMessage.PrepareCb{
	private ImgMessage mImgMessageData = null;

	private ProgressBar mSpinner = null;
	private ImgMessageView mImgMessageView = null;

	private Bitmap mImgBitmap = null;
	private ImgMessage.Background mMsgBg = null;
	private ImgMessage.Shim mMsgShim = null;

	int mScreenWidth = 0;
	int mScreenHeight = 0;
	/**
	 * Custom view that overrides onDraw() to display the bitmap fragments according to the ImgMessage
	 * requirements.
	 */
	private class ImgMessageView extends View implements View.OnTouchListener{
		/**
		 * Constructor override to set the touch listener.
		 * 
		 * @param context The widget context.
		 */
		public ImgMessageView(Context context){
			super(context);
			this.setOnTouchListener(this);
		}
		/**
		 * Constructor override to set the touch listener.
		 * 
		 * @param context The widget context.
		 * @param attrs The widget attributes.
		 */
		public ImgMessageView(Context context, AttributeSet attrs){
			super(context, attrs);
			this.setOnTouchListener(this);
		}
		@Override
		public boolean performClick(){
			return super.performClick();
		}
		@Override
		protected void onLayout(boolean changed, int left, int top, int right, int bottom){
			mScreenWidth = right;
			mScreenHeight = bottom;

			if(mImgBitmap == null){
				BitmapFactory.Options options = new BitmapFactory.Options();
				options.inPreferredConfig = Bitmap.Config.ARGB_8888;
				mImgBitmap = BitmapFactory.decodeFile(mImgMessageData.imgFilename(), options);
			}

			mImgMessageData.init(getResources().getConfiguration().orientation, mScreenWidth, mScreenHeight);

			mMsgBg = mImgMessageData.background();
			mMsgShim = mImgMessageData.shim();
			
			super.onLayout(changed, left, top, right, bottom);
		}
		@Override
		protected void onDraw(Canvas canvas){
			if(mImgBitmap != null){
				if(ImgMessage.MASK_DIMMED.equalsIgnoreCase(mMsgShim.mask())){
					canvas.drawARGB(0x66, 0x0, 0x0, 0x0);
				}else{
					canvas.drawARGB(0x0, 0x0, 0x0, 0x0);
				}
				int orientation = getContext().getResources().getConfiguration().orientation;

				// draw background
				canvas.drawBitmap(mImgBitmap, mMsgBg.imgRect() , mMsgBg.layout(orientation).frame(), null);

				// draw buttons
				int cnt = 0;
				ImgMessage.Button btn = null;
				while(cnt < mImgMessageData.buttonCount()){
					btn = mImgMessageData.button(cnt);
					canvas.drawBitmap(mImgBitmap, btn.imgRect(), btn.layout(orientation).frame(), null);
					++cnt;
				}

			}else{
				canvas.drawARGB(0x0, 0x0, 0x0, 0x0);
			}
		}
		@SuppressLint("ClickableViewAccessibility")
		@Override		
		public boolean onTouch(View v, MotionEvent event) {
			switch (event.getAction()){
			case MotionEvent.ACTION_UP:{
				if(mImgMessageData != null){
					ImgMessage.Action action = null;
					int orientation = getContext().getResources().getConfiguration().orientation;

					// if the touch is on the popup then test buttons
					if(mMsgBg.layout(orientation).frame().contains((int)event.getX(), (int)event.getY())){
						int cnt = 0;
						ImgMessage.Button btn = null;
						while((action == null) && (cnt < mImgMessageData.buttonCount())){
							btn = mImgMessageData.button(cnt);
							if(btn.layout(orientation).frame().contains((int)event.getX(), (int)event.getY())){
								action = btn.action(orientation);
							}
							++cnt;
						}
						
					}else{
						// touch is outside the popup so use shim action
						action = mMsgShim.action();
					}

					performAction(action);
				}
				break;
			}
			default:
				break;
			}
			return true;
		}
	}
	/**
	 * Handles an action defined by the Delta DNA message.
	 * 
	 * @param action The action to handle.
	 */
	protected void performAction(ImgMessage.Action action){
		if(action != null){
			if(action.type().equalsIgnoreCase(ImgMessage.ACTION_ACTION)){
				Intent resultIntent = new Intent();
				resultIntent.putExtra(DDNA.EXTRA_DDNAVALUE, action.value());
				if(mImgMessageData.parameters() != null){
					resultIntent.putExtra(DDNA.EXTRA_DDNAPARAMS, mImgMessageData.parameters().toString());
				}

				setResult(Activity.RESULT_OK, resultIntent);
				finish();

			}else if(action.type().equalsIgnoreCase(ImgMessage.ACTION_DISMISS)){
				mImgMessageData.cleanUp();
				Intent resultIntent = new Intent();

				setResult(Activity.RESULT_CANCELED, resultIntent);
				finish();
			}
		}
	}
	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		RelativeLayout contentLayout = new RelativeLayout(this);
		
		FrameLayout holder = new FrameLayout(this);
		contentLayout.addView(holder, new RelativeLayout.LayoutParams(RelativeLayout.LayoutParams.MATCH_PARENT, RelativeLayout.LayoutParams.MATCH_PARENT));

		mImgMessageView = new ImgMessageView(this);
		holder.addView(mImgMessageView, new LayoutParams(LayoutParams.MATCH_PARENT, LayoutParams.MATCH_PARENT));

		mSpinner = new ProgressBar(this, null, android.R.attr.progressBarStyleLarge);
		mSpinner.setIndeterminate(true);
		mSpinner.setVisibility(View.VISIBLE);
		RelativeLayout.LayoutParams params = new RelativeLayout.LayoutParams(RelativeLayout.LayoutParams.WRAP_CONTENT, RelativeLayout.LayoutParams.WRAP_CONTENT);
		params.addRule(RelativeLayout.CENTER_IN_PARENT);
		contentLayout.addView(mSpinner, params);
		
		setContentView(contentLayout);
		
		String json = getIntent().getStringExtra(DDNA.EXTRA_IMGMESSAGE);
		if(json != null){
			try {
				mImgMessageData = new ImgMessage(json);
				mImgMessageData.prepare(this);
			} catch (JSONException e) {
				Log.d(DDNA.LOGTAG, "Image message JSON is invalid: "+json);
			}
		}
	}
	@Override
	protected void onDestroy(){
		super.onDestroy();
	}
	@Override
	protected void onResume(){
		super.onResume();

	}
	@Override
	protected void onPause(){

		super.onPause();
	}
	@Override
	public void onBackPressed(){
		moveTaskToBack(true);
	}

	@Override
	public void ready(ImgMessage src) {
		runOnUiThread(new Runnable(){
			@Override
			public void run(){
				mSpinner.setVisibility(View.INVISIBLE);
				mImgMessageView.requestLayout();
			}
		});
	}
	@Override
	public void progress(ImgMessage src) {
	}
	@Override
	public void error(ImgMessage src) {
		mImgMessageData.cleanUp();
		setResult(Activity.RESULT_CANCELED, null);
		finish();
	}
}
