package com.deltadna.android.sdk;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
/**
 * Helper class to construct a complex product event.
 */
public class ProductBuilder{
	private JSONObject mRealCurrency;
	private JSONArray mVirtualCurrencies;
	private JSONArray mItems;
	/**
	 * Constructor.
	 */
	public ProductBuilder(){}
	/**
	 * Adds the single real currency to the product.
	 * 
	 * @param currencyType The type of currency.
	 * @param currencyAmount The amount of currency.
	 * 
	 * @return Returns this instance.
	 */
	public ProductBuilder addRealCurrency(String currencyType, int currencyAmount){
		if (mRealCurrency != null){
			throw new IllegalStateException("A Product may only have one real currency");
		}

		mRealCurrency = new JSONObject();
		try {
			mRealCurrency.put("realCurrencyType", currencyType);
			mRealCurrency.put("realCurrencyAmount", currencyAmount);
		} catch (JSONException e) {}

		return this;
	}
	/**
	 * Adds a virtual currency, multiple virtual currencies can be added.
	 * 
	 * @param currencyName The virtual currency name.
	 * @param currencyType The virtual currency type.
	 * @param currencyAmount The virtual currency amount.
	 * 
	 * @return Returns this instance.
	 */
	public ProductBuilder addVirtualCurrency(String currencyName, String currencyType, int currencyAmount){
		if (mVirtualCurrencies == null){
			mVirtualCurrencies = new JSONArray();
		}

		JSONObject newVCurrency = new JSONObject();
		try{
			newVCurrency.put("virtualCurrencyName", currencyName);
			newVCurrency.put("virtualCurrencyType", currencyType);
			newVCurrency.put("virtualCurrencyAmount", currencyAmount);
			
			newVCurrency = new JSONObject().put("virtualCurrency", newVCurrency);
		}catch(JSONException e){}

		mVirtualCurrencies.put(newVCurrency);

		return this;
	}
	/**
	 * Adds an item to the product, multiple items may be added.
	 * 
	 * @param itemName The item name.
	 * @param itemType The item type.
	 * @param itemAmount The item amount.
	 * 
	 * @return Returns this instance.
	 */
	public ProductBuilder addItem(String itemName, String itemType, int itemAmount){
		if (mItems == null){
			mItems = new JSONArray();
		}

		JSONObject newItem = new JSONObject();
		try{
			newItem.put("itemName", itemName);
			newItem.put("itemType", itemType);
			newItem.put("itemAmount", itemAmount);
			
			newItem = new JSONObject().put("item", newItem);
		}catch(JSONException e){}

		mItems.put(newItem);

		return this;
	}
	/**
	 * Returns a JSON representation of the EventBuilder.
	 * 
	 * @return A JSON representation of the EventBuilder.
	 */
	public JSONObject toJSON(){
		JSONObject contents = new JSONObject();

		try{
			if (mRealCurrency != null){
				contents.put("realCurrency", mRealCurrency);
			}
			if (mVirtualCurrencies != null){
				contents.put("virtualCurrencies", mVirtualCurrencies);
			}
			if (mItems != null){
				contents.put("items", mItems);
			}
		}catch(JSONException e){}

		return contents;
	}
}
