package com.deltadna.android.sdk.helpers;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

import com.deltadna.android.sdk.DDNA;

import android.util.Log;

/**
 * The Engage archive holds previously requested engage responses. The responses
 * can be saved to disk.
 */
public class EngageArchive{
	static private final String FILENAME = "ENGAGEMENTS";

	private HashMap<String, String> mTable = new HashMap<String, String>();
	private Object mLock = new Object();
	private String mPath;

	/**
	 * Creates a new EnagageArchive, loading any previously saved Engagements from
	 * a file at <param name="path">. 
	 * 
	 * @param path The data file path.
	 */
	public EngageArchive(String path){
		mPath = path;
		load(mPath);
	}
	/**
	 * Returns true if the archive contains a previous response for the
	 * decision point.
	 * 
	 * @param decisionPoint Decision point name.
	 * 
	 * @return TRUE if store contains a response, FALSE otherwise.
	 */
	public boolean contains(String decisionPoint){
		Log.v(DDNA.LOGTAG, "Does Engage contain " + decisionPoint);
		return mTable.containsKey(decisionPoint);
	}
	/**
	 * Gets a decision point.
	 * 
	 * @param decisionPoint Decision point name.
	 * 
	 * @return The data on success, null otherwise.
	 */
	public String get(String decisionPoint){
		return mTable.get(decisionPoint);
	}
	/**
	 * Puts a decision point in the archive.
	 * 
	 * @param decisionPoint The decision point name.
	 * @param value The decision point data.
	 */
	public void put(String decisionPoint, String value){
		mTable.put(decisionPoint, value);
	}
	/**
	 * Loads an existing archive from disk.
	 * 
	 * @param path The path to load from.
	 */
	private void load(String path){
		synchronized(mLock){
			FileInputStream fs = null;

			try{
				File file = new File(path, FILENAME);

				Log.d(DDNA.LOGTAG, "Loading Engage from " + file.getAbsolutePath());

				if(file.exists()){
					fs = new FileInputStream(file);

					String key = null;
					String value = null;
					int read = 0;
					byte[] length = new byte[4];
					int valueLength;
					byte[] valueField;
					while(fs.read(length, 0, length.length) > 0){
						valueLength = Utils.toInt32(length);
						valueField = new byte[valueLength];
						fs.read(valueField, 0, valueField.length);
						if(read % 2 == 0){
							key = new String(valueField, "UTF-8");
						}else{
							value = new String(valueField, "UTF-8");
							mTable.put(key, value);
						}
						read++;
					}
				}
			}catch (Exception e){
				Log.w(DDNA.LOGTAG, "Unable to load Engagement archive: " + e.getMessage());

			}finally{
				if(fs != null){
					try {
						fs.close();
					} catch (IOException e) {}
				}
			}
		}
	}
	/**
	 * Save the archive to disk.
	 */
	public void save(){
		synchronized(mLock){
			FileOutputStream fs = null;

			try{
				File dir = new File(mPath);

				if(!dir.exists()){
					dir.mkdirs();
				}

				ByteArrayOutputStream bytes = new ByteArrayOutputStream();
				byte[] keyBytes = null;
				byte[] keyLenBytes = null;
				byte[] valueBytes = null;
				byte[] valueLenBytes = null;
				String key = null;
				String value = null;
				Set<String> keys = mTable.keySet();
				Iterator<String> iter = keys.iterator();
				while(iter.hasNext()){
					key = iter.next();
					value = mTable.get(key);
					
					keyBytes = key.getBytes();
					keyLenBytes = Utils.toBytes(keyBytes.length);
					valueBytes = value.getBytes();
					valueLenBytes = Utils.toBytes(valueBytes.length);

					bytes.write(keyLenBytes);
					bytes.write(keyBytes);
					bytes.write(valueLenBytes);
					bytes.write(valueBytes);
				}

				File file = new File(dir.getAbsolutePath(), FILENAME);
				fs = new FileOutputStream(file);

				byte[] byteArray = bytes.toByteArray();
				
				fs.write(byteArray, 0, byteArray.length);
				
			}catch (Exception e){
				Log.w(DDNA.LOGTAG, "Unable to save Engagement archive: " + e.getMessage());
				
			}finally{
				if(fs != null){
					try {
						fs.close();
					} catch (IOException e) {}
				}
			}
		}
	}
	/**
	 * Clears the archive.
	 */
	public void clear(){
		synchronized(mLock){
			mTable.clear();
		}
	}
}
