package com.deltadna.android.sdk.net;

import java.util.Locale;
import java.util.Vector;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.http.HttpVersion;
import org.apache.http.conn.params.ConnManagerParams;
import org.apache.http.conn.params.ConnPerRouteBean;
import org.apache.http.conn.scheme.PlainSocketFactory;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.tsccm.ThreadSafeClientConnManager;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.HttpConnectionParams;
import org.apache.http.params.HttpProtocolParams;
import org.apache.http.protocol.BasicHttpContext;
import org.apache.http.protocol.HttpContext;
import org.apache.http.protocol.SyncBasicHttpContext;
import org.json.JSONException;
import org.json.JSONObject;

import android.content.Context;
import android.net.ConnectivityManager;
import android.net.NetworkInfo;
import android.util.Log;

import com.deltadna.android.sdk.DDNA;
import com.deltadna.android.sdk.helpers.ClientInfo;
import com.deltadna.android.sdk.helpers.EventStore;
import com.deltadna.android.sdk.helpers.MD5;
import com.deltadna.android.sdk.helpers.NotStartedException;
import com.deltadna.android.sdk.helpers.Settings;
import com.deltadna.android.sdk.helpers.Utils;

/**
 * The net manager handles all network requests in a queue. Requests are
 * all threaded to allow asynchronous operation.
 * 
 * A generic static file fetch is also implemented for static config files
 * and images.
 */


public class DDNANetManager implements Runnable{
	static private DDNANetManager smInst = null;

	static public int NET_MAXREQUESTS = 10;
	static public int NET_REQTIMEOUT = 5000;
	static public int NET_REQMAXCONNPERROUTE = 10;
	static public int NET_REQBUFFERSIZE = 32768;

	private Thread mThread = null;
	private boolean mRunning = false;

	private DDNANetReq mUploadReq = null;
	private DDNANetReq mEngageReq = null;
	private DDNANetReq mImgMsgReq = null;
	private Vector<DDNANetReq> mFileRequests = new Vector<DDNANetReq>();

	private Vector<DDNANetReq> mDeadRequests = new Vector<DDNANetReq>(); //!< not a great pattern without a request queue but needs must to call join on cancelled requests.

	protected ReentrantLock mLock = new ReentrantLock();

	private DefaultHttpClient mHTTPClient = null;
	private HttpContext mHTTPContext = null;

	private long mUploadRepeatRate = 1000;
	private long mUploadStartDelay = 1000;
	private long mThreadStartTime = 0;

	/**
	 * Default constructor.
	 */
	private DDNANetManager(){
		BasicHttpParams httpParams = new BasicHttpParams();

		ConnManagerParams.setTimeout(httpParams, NET_REQTIMEOUT);
		ConnManagerParams.setMaxConnectionsPerRoute(httpParams, new ConnPerRouteBean(NET_REQMAXCONNPERROUTE));
		ConnManagerParams.setMaxTotalConnections(httpParams, NET_MAXREQUESTS);

		HttpConnectionParams.setSoTimeout(httpParams, NET_REQTIMEOUT);
		HttpConnectionParams.setConnectionTimeout(httpParams, NET_REQTIMEOUT);
		HttpConnectionParams.setTcpNoDelay(httpParams, true);
		HttpConnectionParams.setSocketBufferSize(httpParams, NET_REQBUFFERSIZE);

		HttpProtocolParams.setVersion(httpParams, HttpVersion.HTTP_1_1);
		HttpProtocolParams.setUserAgent(httpParams, "deltaDNA Android SDK: " + DDNA.inst().clientVersion());

		SchemeRegistry schemeRegistry = new SchemeRegistry();
		schemeRegistry.register(new Scheme("http", PlainSocketFactory.getSocketFactory(), 80));
		schemeRegistry.register(new Scheme("https", SSLSocketFactory.getSocketFactory(), 443));
		ThreadSafeClientConnManager cm = new ThreadSafeClientConnManager(httpParams, schemeRegistry);

		mHTTPContext = new SyncBasicHttpContext(new BasicHttpContext());
		mHTTPClient = new DefaultHttpClient(cm, httpParams);
	}
	/**
	 * Gets the http client.
	 * 
	 * @return Gets the http client of the manager.
	 */
	public DefaultHttpClient httpClient(){
		return mHTTPClient;
	}
	/**
	 * Gets the http context.
	 * 
	 * @return Gets the http context of the manager.
	 */
	public HttpContext httpContext(){
		return mHTTPContext;
	}
	/**
	 * Tests if the manager is currently uploading events.
	 * 
	 * @return TRUE if events are being uploaded, FALSE otherwise.
	 */
	public boolean isUploading(){
		return (mUploadReq != null);
	}
	/**
	 * Internal method that uploads the events if any exist in the EventStore. This internal method
	 * does not lock, it relies on the background thread to lock the manager before calling.
	 * 
	 * @param cb The optional response callback.
	 * 
	 * @return The current upload request, null on error.
	 * @throws NotStartedException 
	 */
	private DDNANetReq uploadEvents_internal(DDNANetReqCb cb) throws NotStartedException{
		if(DDNA.inst().started()){
			if(mUploadReq == null){
				// Swap over event queue.
				EventStore eventStore = DDNA.inst().eventStore();
				eventStore.swap();	

				// Create bulk event message to post.
				Vector<String> events = eventStore.read();

				if(events.size() > 0){
					Log.d(DDNA.LOGTAG, "Starting event upload");

					StringBuilder body = new StringBuilder("{\"eventList\":[");
					int cnt = 0;
					while(cnt < events.size()){
						body.append(events.elementAt(cnt));

						++cnt;
						if(cnt < events.size()){
							body.append(",");
						}
					}
					body.append("]}");

					String url;
					if (DDNA.inst().hashSecret() != null){
						String md5Hash = MD5.getMd5String(body.toString() + DDNA.inst().hashSecret());
						url = DDNANetManager.formatURI(Settings.COLLECT_HASH_URL_PATTERN, DDNA.inst().collectURL(), DDNA.inst().environmentKey(), md5Hash);
					}else{
						url = DDNANetManager.formatURI(Settings.COLLECT_URL_PATTERN, DDNA.inst().collectURL(), DDNA.inst().environmentKey());
					}


					mUploadReq = new DDNANetReq(this, DDNANetReq.REQ_ACTION_POST, url, body.toString(), new SDKUploadCb(cb));
					mUploadReq.send();
				}

			}else{
				Log.w(DDNA.LOGTAG, "Event upload already in progress, aborting");
			}

		}else{
			throw new NotStartedException("You must first start the SDK via the startSDK method");
		}

		return mUploadReq;
	}
	/**
	 * Uploads waiting events to our Collect service.  By default this is called automatically in the 
	 * background.  If you disable auto uploading via <see cref="Settings.BackgroundEventUpload"/> you
	 * will need to call this method yourself periodically.
	 * 
	 * @param cb Optional user callback to notify on network events.
	 * @throws NotStartedException
	 */
	public DDNANetReq uploadEvents(DDNANetReqCb cb) throws NotStartedException{
		mLock.lock();

		clearDead();

		try{
			mUploadReq = uploadEvents_internal(cb);
		}finally{
			mLock.unlock();
		}

		return mUploadReq;
	}
	/**
	 * Internal callback for upload status.
	 */
	private class SDKUploadCb implements DDNANetReqCb{
		private DDNANetReqCb mUserCb = null;
		/**
		 * Constructor.
		 * 
		 * @param userCb Optional user callback.
		 */
		public SDKUploadCb(DDNANetReqCb userCb){
			mUserCb = userCb;
		}
		@Override
		public void done(DDNANetReq req) {
			Log.d(DDNA.LOGTAG, "Event upload successful");
			((EventStore)DDNA.inst().eventStore()).clearOutfile();

			if(mUserCb != null){
				mUserCb.done(req);
			}
			mUploadReq = null;
		}
		@Override
		public void progress(DDNANetReq req) {
			if(mUserCb != null){
				mUserCb.progress(req);
			}
		}
		@Override
		public void failed(DDNANetReq req) {
			Log.w(DDNA.LOGTAG, "Event upload failed - ");
			if(req.httpResponseCode() == 400){
				((EventStore)DDNA.inst().eventStore()).clearOutfile();
				Log.w(DDNA.LOGTAG, "Unrecoverable, data lost.");
			}else{
				Log.w(DDNA.LOGTAG, "Try again later.");
			}
			if(mUserCb != null){
				mUserCb.failed(req);
			}
			mUploadReq = null;
		}
	}
	/**
	 * Makes an Engage request.  The result of the engagement will be passed as a dictionary object to your callback method.
	 * 
	 * @param decisionPoint The decision point the request is for, must match the String in Portal.
	 * @param engageParams Additional parameters for the engagement.
	 * @param cb Optional user callback to notify on network events.
	 * @throws NotStartedException
	 */
	public DDNANetReq engageRequest(String decisionPoint, JSONObject engageParams, DDNANetReqCb cb) throws NotStartedException{
		if(DDNA.inst().started()){
			if(!Utils.isNullOrEmpty(DDNA.inst().engageURL())){
				if(!Utils.isNullOrEmpty(decisionPoint)){
					mLock.lock();

					clearDead();

					if (mEngageReq == null){
						Log.d(DDNA.LOGTAG, "Starting engagement for '"+decisionPoint+"'");	

						JSONObject engageRequest = new JSONObject();
						try {
							engageRequest.put("userID", DDNA.inst().userID());
							engageRequest.put("decisionPoint", decisionPoint);
							engageRequest.put("sessionID", DDNA.inst().sessionID());
							engageRequest.put("version", Settings.ENGAGE_API_VERSION);
							engageRequest.put("sdkVersion", Settings.SDK_VERSION);
							engageRequest.put("platform", DDNA.inst().platform());
							engageRequest.put("timezoneOffset", ClientInfo.timezoneOffset());

							if (ClientInfo.locale() != null){
								engageRequest.put("locale", ClientInfo.locale());
							}

							if (engageParams != null){
								engageRequest.put("parameters", engageParams);
							}
						} catch (JSONException e) {}

						String engageJSON = engageRequest.toString();

						String url;
						if (DDNA.inst().hashSecret() != null){
							String md5Hash = MD5.getMd5String(engageJSON + DDNA.inst().hashSecret());
							url = DDNANetManager.formatURI(Settings.ENGAGE_HASH_URL_PATTERN, DDNA.inst().engageURL(), DDNA.inst().environmentKey(), md5Hash);
						}else{
							url = DDNANetManager.formatURI(Settings.ENGAGE_URL_PATTERN, DDNA.inst().engageURL(), DDNA.inst().environmentKey());
						}

						mEngageReq = new DDNANetReq(this, DDNANetReq.REQ_ACTION_POST, url, engageJSON, new SDKEngageCb(decisionPoint, cb));
						mEngageReq.send();

					}else{
						Log.w(DDNA.LOGTAG, "Engage request already in progress, aborting");
					}

					mLock.unlock();

				}else{
					Log.w(DDNA.LOGTAG, "No decision point set, can not make engagement.");
				}
			}else{
				Log.w(DDNA.LOGTAG, "Engage URL not configured, can not make engagement.");
			}
		}else{
			throw new NotStartedException("You must first start the SDK via the startSDK method");
		}

		return mEngageReq;
	}
	/**
	 * Internal callback for Engage request.
	 */
	private class SDKEngageCb implements DDNANetReqCb{
		private String mDecisionPoint = null;
		private DDNANetReqCb mUserCb = null;
		/**
		 * Constructor.
		 * 
		 * @param decisionPoint Decision point string name.
		 * @param userCb Optional user callback.
		 */
		public SDKEngageCb(String decisionPoint, DDNANetReqCb userCb){
			mDecisionPoint = decisionPoint;
			mUserCb = userCb;
		}
		@Override
		public void done(DDNANetReq req) {
			String resp = req.responseData().toString();
			Log.d(DDNA.LOGTAG, "Using live engagement: " + resp);

			DDNA.inst().engageArchive().put(mDecisionPoint, resp);

			if(mUserCb != null){
				mUserCb.done(req);
			}
			mEngageReq = null;
		}
		@Override
		public void progress(DDNANetReq req) {
			if(mUserCb != null){
				mUserCb.progress(req);
			}
		}
		@Override
		public void failed(DDNANetReq req) {
			String cachedPoint = DDNA.inst().engageArchive().get(mDecisionPoint);

			if(cachedPoint != null){
				Log.w(DDNA.LOGTAG, "Engage request failed, use cached response.");
				try {
					JSONObject json = new JSONObject(cachedPoint);
					json.put("isCachedResponse", true);
					req.setResponseData(json.toString().getBytes());
				} catch (JSONException e) {
					e.printStackTrace();
				}
			}else{
				Log.w(DDNA.LOGTAG, "Engage request failed");
			}

			if(mUserCb != null){
				mUserCb.failed(req);
			}
			mEngageReq = null;
		}
	}
	/**
	 * Requests an image based Engagement for popping up on screen.  This is a convience around requestEngagement 
	 * that loads the image resource automatically from the original engage request.  Register a function with the 
	 * Popup's AfterLoad event to be notified when the image has be been downloaded from our server.
	 * 
	 * @param decisionPoint The decision point the request is for, must match the String in Portal.
	 * @param engageParams Additional parameters for the engagement.
	 * @param cb Optional user callback to notify on network events.
	 * @throws NotStartedException
	 */
	public DDNANetReq requestImageMessage(String decisionPoint,
			JSONObject engageParams,
			DDNANetReqCb cb) throws NotStartedException{
		if(DDNA.inst().started()){
			if (mImgMsgReq == null){
				mLock.lock();

				clearDead();

				if (mImgMsgReq == null){
					Log.d(DDNA.LOGTAG, "Starting engagement for '"+decisionPoint+"'");	

					JSONObject engageRequest = new JSONObject();
					try{
						engageRequest.put("userID", DDNA.inst().userID());
						engageRequest.put("decisionPoint", decisionPoint);
						engageRequest.put("sessionID", DDNA.inst().sessionID());
						engageRequest.put("version", Settings.ENGAGE_API_VERSION);
						engageRequest.put("sdkVersion", Settings.SDK_VERSION);
						engageRequest.put("platform", DDNA.inst().platform());
						engageRequest.put("timezoneOffset", ClientInfo.timezoneOffset());

						if (ClientInfo.locale() != null){
							engageRequest.put("locale", ClientInfo.locale());
						}

						if (engageParams != null){
							engageRequest.put("parameters", engageParams);
						}
					}catch(JSONException e){}

					String engageJSON = engageRequest.toString();

					String url;
					if (DDNA.inst().hashSecret() != null){
						String md5Hash = MD5.getMd5String(engageJSON + DDNA.inst().hashSecret());
						url = DDNANetManager.formatURI(Settings.ENGAGE_HASH_URL_PATTERN, DDNA.inst().engageURL(), DDNA.inst().environmentKey(), md5Hash);
					}else{
						url = DDNANetManager.formatURI(Settings.ENGAGE_URL_PATTERN, DDNA.inst().engageURL(), DDNA.inst().environmentKey());
					}

					mImgMsgReq = new DDNANetReq(this, DDNANetReq.REQ_ACTION_POST, url, engageJSON, new SDKImgReqCb(cb));
					mImgMsgReq.send();

				}else{
					Log.w(DDNA.LOGTAG, "Engage request already in progress, aborting");
				}

				mLock.unlock();

			}else{
				Log.w(DDNA.LOGTAG, "Image message request already in progress, aborting");
			}

		}else{
			throw new NotStartedException("You must first start the SDK via the startSDK method");
		}

		return mImgMsgReq;
	}
	/**
	 * Internal callback for image message request.
	 */
	private class SDKImgReqCb implements DDNANetReqCb{
		private DDNANetReqCb mUserCb = null;
		/**
		 * Constructor.
		 * 
		 * @param userCb Optional user callback.
		 */
		public SDKImgReqCb(DDNANetReqCb userCb){
			mUserCb = userCb;
		}
		@Override
		public void done(DDNANetReq req) {
			Log.d(DDNA.LOGTAG, "Using image message: " + req.responseData().toString());

			if(mUserCb != null){
				mUserCb.done(req);
			}
			mEngageReq = null;
		}
		@Override
		public void progress(DDNANetReq req) {
			if(mUserCb != null){
				mUserCb.progress(req);
			}
		}
		@Override
		public void failed(DDNANetReq req) {
			Log.w(DDNA.LOGTAG, "Image request failed");

			if(mUserCb != null){
				mUserCb.failed(req);
			}
			mEngageReq = null;
		}
	}
	/**
	 * Fetches static content from the given url. An optional destination file can be provided.
	 * 
	 * @param url The location of the data to fetch.
	 * @param destFile The destination file path and name.
	 * @param cb Optional request status callback.
	 * 
	 * @return A pointer to the request object, null on error.
	 */
	public DDNANetReq fetchStatic(String url, String destFile, DDNANetReqCb cb){
		DDNANetReq req = null;

		if(url != null){
			mLock.lock();

			clearDead();

			req = new DDNANetReq(this, url, destFile, cb);
			mFileRequests.add(req);
			req.send();

			mLock.unlock();
		}

		return req;
	}
	/**
	 * Cancels all current active requests and kills anything in the queue.
	 */
	public void cancelAll(){
		mLock.lock();

		// find the request
		DDNANetReq temp = null;

		// check the active request queue
		while(mFileRequests.size() > 0){
			temp = mFileRequests.get(0);
			temp.cancel();
			mDeadRequests.add(temp);
			mFileRequests.remove(0);
		}

		if(mUploadReq != null){
			mUploadReq.cancel();
			mDeadRequests.add(mUploadReq);
		}
		if(mEngageReq != null){
			mEngageReq.cancel();
			mDeadRequests.add(mEngageReq);
		}
		if(mImgMsgReq != null){
			mImgMsgReq.cancel();
			mDeadRequests.add(mImgMsgReq);
		}

		clearDead();

		mLock.unlock();
	}
	/**
	 * Cancels the specified request.
	 * 
	 * @param req The request to cancel.
	 * 
	 * @return TRUE if the request was canceled, FALSE otherwise.
	 */
	public boolean cancelReq(DDNANetReq req){
		boolean result = false;

		mLock.lock();

		// find the request
		DDNANetReq temp = null;

		// check the active request queue
		int cnt = 0;
		while(cnt < mFileRequests.size()){
			temp = mFileRequests.get(cnt);

			if(temp == req){
				mFileRequests.remove(cnt);
				cnt = mFileRequests.size();
			}else{
				temp = null;
			}

			++cnt;
		}

		if(mUploadReq == req){
			temp = mUploadReq;
		}
		if(mEngageReq == req){
			temp = mEngageReq;
		}
		if(mImgMsgReq == req){
			temp = mImgMsgReq;
		}

		// cancel and destroy
		if(temp != null){
			temp.cancel();
			mDeadRequests.add(temp);
			result = true;
		}

		clearDead();

		mLock.unlock();

		return result;
	}

	/**
	 * Tests if the network manager is uploading data in the background.
	 * 
	 * @return TRUE if the background thread is active, FALSE otherwise.
	 */
	public boolean isRunning(){
		return mRunning;
	}
	/**
	 * Starts the background data upload thread.
	 * 
	 * @param startDelay The delay before the thread starts.
	 * @param uploadRepeatRate The delay between uploads.
	 * 
	 * @return TRUE if the background thread is running, FALSE otherwise.
	 */
	public boolean start(int startDelay, int uploadRepeatRate){
		boolean result = false;

		if(mThread == null){
			mThread = new Thread(this);
		}

		if(!mThread.isAlive()){
			mUploadRepeatRate = uploadRepeatRate * 1000;
			mUploadStartDelay = startDelay * 1000;
			mRunning = true;
			mThreadStartTime = System.currentTimeMillis();
			mThread.start();

		}else{
			result = true;
		}

		return result;
	}
	/**
	 * Stops the background upload thread.
	 */
	public void stop(){
		if((mThread != null) && mThread.isAlive()){
			mRunning = false;
			try {
				mThread.interrupt();
				mThread.join();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}finally{
				mThread = null;
			}
		}
	}
	@Override
	public void run() {
		while(mRunning){
			if((System.currentTimeMillis() - mThreadStartTime) > mUploadStartDelay){
				if(mLock.tryLock()){
					clearDead();

					try{
						uploadEvents_internal(null);
					}catch(NotStartedException e){
					}finally{
						mLock.unlock();
					}
				}
			}

			try {
				Thread.sleep(mUploadRepeatRate);
			} catch (InterruptedException e) {
				mRunning = false;
			}
		}

	}
	/**
	 * Handler called by requests when it has completed.
	 * 
	 * @param req The request that has completed.
	 */
	protected void reqComplete(DDNANetReq req){
		req.processUserCallback();

		// wait for lock on API to remove the response
		mLock.lock();

		mFileRequests.remove(req);

		if(mUploadReq == req){
			mUploadReq = null;
		}
		if(mEngageReq == req){
			mEngageReq = null;
		}
		if(mImgMsgReq == req){
			mImgMsgReq = null;
		}

		mDeadRequests.add(req);

		mLock.unlock();
	}
	/**
	 * Handler called by a request when it has failed.
	 * 
	 * @param req The request that has failed.
	 */
	protected void reqFailed(DDNANetReq req){
		req.processUserCallback();

		mLock.lock();

		mFileRequests.remove(req);

		if(mUploadReq == req){
			mUploadReq = null;
		}
		if(mEngageReq == req){
			mEngageReq = null;
		}
		if(mImgMsgReq == req){
			mImgMsgReq = null;
		}

		mDeadRequests.add(req);

		mLock.unlock();
	}
	/**
	 * Formats a uri string for delta dna.
	 * 
	 * @param uriPattern The pattern to use.
	 * @param apiHost The host uri.
	 * @param envKey The environment key.
	 * 
	 * @return The formatted uri.
	 */
	static public String formatURI(String uriPattern, String apiHost, String envKey){
		return formatURI(uriPattern, apiHost, envKey, null);
	}
	/**
	 * Formats a uri string for delta dna.
	 * 
	 * @param uriPattern The pattern to use.
	 * @param apiHost The host uri.
	 * @param envKey The environment key.
	 * @param hash Optional hash value.
	 * 
	 * @return The formatted uri.
	 */
	static public String formatURI(String uriPattern, String apiHost, String envKey, String hash) {
		String uri = uriPattern.replace("{host}", apiHost);
		uri = uri.replace("{env_key}", envKey);
		if(hash != null){
			uri = uri.replace("{hash}", hash);
		}
		return uri;
	}
	/**
	 * Validates a uri for delta dna, must use http:// protocol.
	 * 
	 * @param url The url to test.
	 * 
	 * @return The url or a new url formatted to http.
	 */
	static public String validateURL(String url) {
		if (!url.toLowerCase(Locale.getDefault()).startsWith("http://")) {
			url = "http://" + url;
		}
		return url;
	}
	/**
	 * Tests if the device is connected to the internet on some medium.
	 * 
	 * @param context A context to use.
	 * 
	 * @return TRUE if connected, FALSE otherwise.
	 */
	static public boolean isConnectedToInternet(Context context){
		boolean result = false;

		ConnectivityManager connectivity = (ConnectivityManager)context.getSystemService(Context.CONNECTIVITY_SERVICE);
		if(connectivity != null){
			NetworkInfo[] info = connectivity.getAllNetworkInfo();
			if (info != null){
				int cnt = 0;
				while(cnt < info.length){
					if (info[cnt].getState() == NetworkInfo.State.CONNECTED){
						result = true;
						cnt = info.length;
					}
					++cnt;
				}
			}
		}

		return result;
	}
	/**
	 * Tests if the device is currently using a WiFi internet connection.
	 * 
	 * @param context A Context to use.
	 * 
	 * @return TRUE if connected via WiFi, FALSE otherwise.
	 */
	static public boolean isConnectedToWiFi(Context context){
		boolean result = false;

		ConnectivityManager connectivity = (ConnectivityManager)context.getSystemService(Context.CONNECTIVITY_SERVICE);
		if(connectivity != null){
			NetworkInfo[] info = connectivity.getAllNetworkInfo();
			if (info != null){
				int cnt = 0;
				while(cnt < info.length){
					if (info[cnt].getTypeName().equalsIgnoreCase("WIFI")){
						if (info[cnt].getState() == NetworkInfo.State.CONNECTED){
							result = true;
							cnt = info.length;
						}
					}
					++cnt;
				}
			}
		}

		return result;
	}
	/**
	 * Clears dead request queue.
	 */
	protected void clearDead(){
		DDNANetReq temp = null;

		while(mDeadRequests.size() > 0){
			temp = mDeadRequests.get(0);
			temp.cleanUp();
			mDeadRequests.remove(0);
		}
	}
	/**
	 * Cleans up manager resources.
	 */
	public void onDestroy(){
		cancelAll();
		clearDead();
	}
	/**
	 * Implements the singleton pattern for the network manager. The single instance
	 * may only be obtained through this call.
	 * 
	 * @return The singleton DDNANetManager instance.
	 */
	static public DDNANetManager inst(){
		if(smInst == null){
			smInst = new DDNANetManager();
		}
		return smInst;
	}
}
