package com.deltadna.android.sdk.net;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.StatusLine;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.message.BasicHeader;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.deltadna.android.sdk.DDNA;
import com.deltadna.android.sdk.helpers.Settings;

/**
 * A threaded network request.
 */
public class DDNANetReq implements Runnable{
	static public final int BODY_JSON = 0;
	static public final int BODY_FORM = 1;
	static public final int BODY_URL = 2;

	static public final int REQ_STATUS_UNKNOWN	 	= 0;		//!< Network request state unknown.
	static public final int REQ_STATUS_READY 		= 1;		//!< Network request state ready to send.
	static public final int REQ_STATUS_ACTIVE 		= 2;		//!< Network request state active.
	static public final int REQ_STATUS_CANCELLED 	= 3;		//!< Network request state inactive user cancel.
	static public final int REQ_STATUS_SUCCEEDED 	= 4;		//!< Network request state inactive succeeded.
	static public final int REQ_STATUS_FAILED 		= 5;		//!< Network request state inactive failed.

	static public final int REQ_ACTION_GET 			= 1;		//!< 
	static public final int REQ_ACTION_POST 		= 2;		//!< 
	static public final int REQ_ACTION_PUT 			= 3;		//!< 
	static public final int REQ_ACTION_DELETE 		= 4;		//!< 

	private Thread 				mThread = null;
	private HttpUriRequest 		mRequest = null;

	private DDNANetManager		mNetManager = null;			//!< The parent network manager.
	private DDNANetReqCb 		mCb = null;

	private int 				mStatus = REQ_STATUS_UNKNOWN;
	private int					mAction = REQ_ACTION_GET;	//!< The http action of the request.
	private int					mRetries = DDNA.inst().settings().httpRequestMaxRetries();

	private String				mUrl = null;						//!< The root url of the request.
	private String				mBody = null;						//!< The list of parameters of the request.

	private boolean 			mStaticDownload = false;
	private long				mExpectedLength = -1;
	private long				mBytesReceived = 0;
	private HttpURLConnection	mStaticConnection = null;

	private String				mDestFile = null;			//!< The destination filename for response data.

	private int					mHttpResponseCode = 0;		//!< The HTTP response code.

	private ByteArrayOutputStream		mResponseData = null;//!< Memory buffer of response data.
	private Object				mJsonResp = null;

	protected ReentrantLock 	mLock = new ReentrantLock();
	/**
	 * Constructor to build a static download request.
	 * 
	 * @param manager The owning network manager.
	 * @param action The HTTP action.
	 * @param url The target url.
	 * @param destFile The destination file.
	 * @param cb
	 */
	protected DDNANetReq(DDNANetManager manager, String url, String destFile, DDNANetReqCb cb){
		mNetManager = manager;
		mUrl = url;
		mDestFile = destFile;
		mCb = cb;
		mStaticDownload = true;

		mStatus = REQ_STATUS_READY;
	}
	/**
	 * Constructor to make a request.
	 * 
	 * @param manager The owning network manager.
	 * @param action The HTTP action.
	 * @param url The target url.
	 * @param body The optional body.
	 * @param cb Optional status callback.
	 */
	protected DDNANetReq(DDNANetManager manager, int action, String url, String body, DDNANetReqCb cb){
		mNetManager = manager;
		mAction = action;
		mUrl = url;
		mCb = cb;
		mBody = body;

		mStatus = REQ_STATUS_READY;
	}
	/**
	 * Constructor to make a request and store result in given destination file.
	 * 
	 * @param manager The owning network manager.
	 * @param action The HTTP action.
	 * @param url The target url.
	 * @param body The optional body.
	 * @param destFile The destination file.
	 * @param cb Optional status callback.
	 */
	protected DDNANetReq(DDNANetManager manager, int action, String url, String body, String destFile, DDNANetReqCb cb){
		mNetManager = manager;
		mAction = action;
		mUrl = url;
		mCb = cb;
		mBody = body;
		mDestFile = destFile;

		mStatus = REQ_STATUS_READY;
	}
	/**
	 * The current request status.
	 * 
	 * @return The current request status.
	 */
	public int status(){
		return mStatus;
	}
	/**
	 * Http response code, only valid after requests completes.
	 * 
	 * @return The http response code.
	 */
	public int httpResponseCode(){
		return mHttpResponseCode;
	}
	/**
	 * Expected data length, only valid after requests completes.
	 * 
	 * @return Expected data length, only valid after requests completes.
	 */
	public long expectedLength(){
		return mExpectedLength;
	}
	/**
	 * Bytes read from the response.
	 * 
	 * @return Bytes read from the response.
	 */
	public long bytesReceived(){
		return mBytesReceived;
	}
	/**
	 * The destination filename, if any.
	 * 
	 * @return The destination filename.
	 */
	public String destFile(){
		return mDestFile;
	}
	/**
	 * The response data as a byte array, may be null if data is being
	 * flushed to a file.
	 * 
	 * @return The response data, if any.
	 */
	public ByteArrayOutputStream responseData(){
		return mResponseData;
	}
	/**
	 * Overrides the response data and sets the given data.
	 * 
	 * @param newData New reponse data.
	 */
	public void setResponseData(byte[] newData){
		mJsonResp = null;
		mResponseData = new ByteArrayOutputStream();
		try {
			mResponseData.write(newData);
			mResponseData.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	/**
	 * Gets the response data as a JSON object.
	 * 
	 * @return The response data as a JSON object.
	 */
	public JSONObject jsonObject(){
		if(mJsonResp == null){
			if((mResponseData != null) && (mResponseData.size() > 0)){
				String str = new String(mResponseData.toByteArray());
				try {
					mJsonResp = new JSONObject(str);
				} catch (JSONException e) {
					e.printStackTrace();
				}
			}
		}
		return (JSONObject)mJsonResp;
	}
	/**
	 * Gets the response data as a JSON array.
	 * 
	 * @return The response data as a JSON array.
	 */
	public JSONArray jsonArray(){
		if(mJsonResp == null){
			if((mResponseData != null) && (mResponseData.size() > 0)){
				String str = new String(mResponseData.toByteArray());
				try {
					mJsonResp = new JSONArray(str);
				} catch (JSONException e) {
					e.printStackTrace();
				}
			}
		}
		return (JSONArray)mJsonResp;
	}
	/**
	 * Sends the request.
	 * 
	 * @return TRUE if the request is sent, FALSE otherwise.
	 */
	public boolean send(){
		boolean result = false;

		if(mStatus == REQ_STATUS_READY){
			if(mThread == null){
				mThread = new Thread(this);
			}

			if(!mThread.isAlive()){
				mThread.start();

			}else{
				result = true;
			}
		}

		return result;
	}
	/**
	 * Cancels the request.
	 */
	public void cancel(){
		mStatus = REQ_STATUS_CANCELLED;
	}
	/**
	 * Cleans up the request resources.
	 */
	public void cleanUp(){
		if((mThread != null) && mThread.isAlive()){
			try {
				if(mRequest != null){
					try{
						mRequest.abort();
						mRequest = null;
					} catch (UnsupportedOperationException e) {
						e.printStackTrace();
					}
				}

				if(mStaticConnection != null){
					mStaticConnection.disconnect();
					mStaticConnection = null;
				}

				mThread.join();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}

		if(mStatus == REQ_STATUS_CANCELLED){
			if(mDestFile != null){
				File f = new File(mDestFile);
				f.delete();
			}
			mResponseData = null;
		}
	}
	@Override
	public void run(){
		boolean done = false;

		if(mStaticDownload){
			while(!done){
				if(mStatus == REQ_STATUS_READY){
					try {
						mStatus = REQ_STATUS_ACTIVE;

						URL url = new URL(mUrl);
						mStaticConnection = (HttpURLConnection) url.openConnection();

						mHttpResponseCode = mStaticConnection.getResponseCode();

						if((mHttpResponseCode >= HttpStatus.SC_OK) && (mHttpResponseCode < HttpStatus.SC_MULTIPLE_CHOICES)){
							mExpectedLength = mStaticConnection.getContentLength();

							OutputStream outStream = null;
							if(mDestFile == null){
								mResponseData = new ByteArrayOutputStream();
								outStream = mResponseData;

							}else{
								try {
									outStream = new FileOutputStream(mDestFile);
								} catch (FileNotFoundException e) {
									e.printStackTrace();
									mStaticConnection.disconnect();
									mStaticConnection = null;
									fail();
									done = true;
								}
							}

							if(outStream != null){
								InputStream inStream = mStaticConnection.getInputStream();

								byte data[] = new byte[40960];
								int count;
								while ((count = inStream.read(data)) != -1) {
									if (mStatus != REQ_STATUS_CANCELLED) {
										outStream.write(data, 0, count);
										mBytesReceived += count;

										if (mExpectedLength > 0){
											if(mCb != null){
												mCb.progress(this);
											}
										}
									}
								}
								mStaticConnection.disconnect();
								mStaticConnection = null;

								inStream.close();
								outStream.flush();
								outStream.close();

								if (mStatus != REQ_STATUS_CANCELLED) {
									success();
								}

								done = true;
							}

						}else{
							mStaticConnection.disconnect();

							if(mRetries == 0){
								fail();
								done = true;
							}else{
								mStatus = REQ_STATUS_READY;
								--mRetries;
							}

						}

					} catch (MalformedURLException e) {
						fail();
						done = true;
					} catch (IOException e) {
						mStaticConnection.disconnect();

						if(mRetries == 0){
							fail();
							done = true;
						}else{
							mStatus = REQ_STATUS_READY;
							--mRetries;
						}
					}
				}
			}

		}else{
			while(!done){
				if(mStatus == REQ_STATUS_READY){
					try{
						if(mAction == REQ_ACTION_GET){
							String url = mUrl;
							mRequest = new HttpGet(url);

						}else if(mAction == REQ_ACTION_POST){
							mRequest = new HttpPost(mUrl);
							if(mBody != null){
								DDNAHttpEntity body = new DDNAHttpEntity(BODY_JSON, mBody.getBytes());
								((HttpPost)mRequest).setEntity(body);
							}
						}else if(mAction == REQ_ACTION_DELETE){
							String url = mUrl;
							mRequest = new HttpDelete(url);

						}else if(mAction == REQ_ACTION_PUT){
							mRequest = new HttpPut(mUrl);
							if(mBody != null){
								DDNAHttpEntity body = new DDNAHttpEntity(BODY_JSON, mBody.getBytes());
								((HttpPost)mRequest).setEntity(body);
							}
						}

						mStatus = REQ_STATUS_ACTIVE;
						HttpResponse response = mNetManager.httpClient().execute(mRequest, mNetManager.httpContext());
						StatusLine statusLine = response.getStatusLine();

						mHttpResponseCode = statusLine.getStatusCode();

						if((mHttpResponseCode >= HttpStatus.SC_OK) && (mHttpResponseCode < HttpStatus.SC_MULTIPLE_CHOICES)){
							if(mDestFile == null){
								mResponseData = new ByteArrayOutputStream();
								if(response.getEntity() != null){
									response.getEntity().writeTo(mResponseData);
									mResponseData.close();
								}

							}else{
								try {
									FileOutputStream outStream = new FileOutputStream(mDestFile);

									response.getEntity().writeTo(outStream);

									outStream.flush();
									outStream = null;

								} catch (FileNotFoundException e) {
									e.printStackTrace();
								} catch (IOException e) {
									e.printStackTrace();
								}
							}

							success();

							done = true;

						}else{
							mResponseData = new ByteArrayOutputStream();
							response.getEntity().writeTo(mResponseData);
							mResponseData.close();

							if(mRetries == 0){
								fail();
								done = true;
							}else{
								mStatus = REQ_STATUS_READY;
								--mRetries;
							}
						}

						mRequest = null;

					}catch(IOException ioe){
						if(mRetries == 0){
							fail();
							done = true;
						}else{
							mStatus = REQ_STATUS_READY;
							--mRetries;
						}
					}
				}
			}
		}
	}
	/**
	 * Processes the optional user callback.
	 */
	protected void processUserCallback(){
		mLock.lock();

		if(mCb != null){
			if(mStatus == REQ_STATUS_ACTIVE){
				mCb.progress(this);
			}else if(mStatus == REQ_STATUS_SUCCEEDED){
				mCb.done(this);
			}else if(mStatus == REQ_STATUS_FAILED){
				mCb.failed(this);
			}
		}

		mLock.unlock();
	}
	/**
	 * Request success handler.
	 */
	protected void success(){
		if(mStatus == REQ_STATUS_ACTIVE){
			mLock.lock();

			mStatus = REQ_STATUS_SUCCEEDED;

			mLock.unlock();

			// call the api
			if(mNetManager != null){
				mNetManager.reqComplete(this);
			}
		}
	}
	/**
	 * Request fail handler.
	 */
	protected void fail(){
		if(mStatus == REQ_STATUS_ACTIVE){
			mLock.lock();

			mStatus = REQ_STATUS_FAILED;

			mLock.unlock();

			// call the api
			if(mNetManager != null){
				mNetManager.reqFailed(this);
			}
		}
	}
	/**
	 * Http entity override to send JSON, url encoded or POST data.
	 */
	public class DDNAHttpEntity implements HttpEntity {
		protected File mFile = null;
		protected FileInputStream mFileStream = null;

		protected byte[] mBytes = null;
		protected int mBody;
		/**
		 * Constructor to use a source file as body.
		 * 
		 * @param body The body type.
		 * @param sourceFile The source file body.
		 */
		protected DDNAHttpEntity(int body, String sourceFile){
			mFile = new File(sourceFile);
			if(mFile.exists()){
				try {
					mFileStream = new FileInputStream(mFile);
				} catch (FileNotFoundException e) {
				}
			}
			mBody = body;
		}
		/**
		 * Constructor to use a byte array as body.
		 * 
		 * @param body The body type.
		 * @param bytes The source bytes.
		 */
		protected DDNAHttpEntity(int body, byte[] bytes){
			mBytes = bytes;
			mBody = body;
		}
		@Override
		public void consumeContent() throws IOException {
		}
		@Override
		public InputStream getContent() throws IOException, IllegalStateException {
			if(mFileStream != null){
				return mFileStream;
			}else{
				return new ByteArrayInputStream(mBytes);
			}
		}
		@Override
		public Header getContentEncoding() {
			return null;
		}
		@Override
		public long getContentLength() {
			if(mFileStream != null){
				return mFile.length();
			}else{
				return mBytes.length;
			}
		}
		@Override
		public Header getContentType() {
			if(mBody == BODY_JSON){
				return new BasicHeader("Content-Type", "application/json");

			}else if(mBody == BODY_FORM){
				return new BasicHeader("Content-Type", "multipart/form-data");

			}else{
				return new BasicHeader("Content-Type", "application/x-www-form-urlencoded");
			}
		}
		@Override
		public boolean isChunked() {
			return false;
		}
		@Override
		public boolean isRepeatable() {
			return false;
		}
		@Override
		public boolean isStreaming() {
			return false;
		}
		@Override
		public void writeTo(OutputStream outstream) throws IOException {
			if(mFileStream != null){
				byte[] bytes = new byte[102400];
				int write = mFileStream.read(bytes, 0, 102400);
				while(write > 0){
					outstream.write(bytes);
					write = mFileStream.read(bytes, 0, 102400);
				}

			}else{
				outstream.write(mBytes);
			}
		}
	}
}
